# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.0.3.dev10+g37cb66b.d20230505'
__version_tuple__ = version_tuple = (0, 0, 3, 'dev10', 'g37cb66b.d20230505')
